"""
Request Argument Standardization
"""

import copy
import json
import logging
import re
from collections import UserDict, abc, defaultdict, namedtuple
from functools import partial
from types import MappingProxyType

try:
    from functools import singledispatchmethod
except ImportError:
    from singledispatchmethod import singledispatchmethod

from biothings.utils.common import dotdict, split_ids


class OptionError(ValueError):

    def __init__(self, reason=None, **kwargs):
        super().__init__()
        self.info = {"reason": reason}
        self.info.update(kwargs)

    def simplify(self):
        self.info = {k: v for k, v in self.info.items() if v}


class Converter():
    """
        A generic HTTP request argument processing unit.
        Only perform one level of validation at this moment.
        The strict switch controls the type conversion rules.
    """

    def __init__(self, **kwargs):

        self.type_ = kwargs.get("type", str)
        self.strict = kwargs.get("strict", True)
        translations = kwargs.get("translations", ())

        self.translations = []

        if isinstance(translations, dict):
            translations = translations.items()

        for pattern, repl in translations:
            if hasattr(re, "Pattern") and isinstance(pattern, re.Pattern): # PY3.6 COMPATIBILITY
                self.translations.append(
                    (pattern, repl))
            elif isinstance(pattern, tuple):
                self.translations.append(
                    (re.compile(*pattern), repl))
            elif isinstance(pattern, str):
                self.translations.append(
                    (re.compile(pattern), repl))
            else:  # https://docs.python.org/3/library/re.html#re.compile
                raise TypeError("Invalid Regex Pattern.")

    def __call__(self, value, to_type):
        return self.convert_to(value, to_type)

    @classmethod
    def subclasses(cls, kwargs):
        for kls in cls.__subclasses__():
            name = kls.__name__[:-len("ArgCvter")]
            yield name.lower(), kls(**kwargs)

    def convert(self, value):
        return self.convert_to(value, self.type_)

    def convert_to(self, value, to_type):

        # default implementation
        # only works for strings
        assert isinstance(value, str)

        if to_type is None:
            to_type = self.type_

        if to_type is str:
            return value  # pass through

        if to_type is bool:
            return self.str_to_bool(value)

        if to_type is int:
            return self.str_to_int(value)

        if to_type in (list, tuple, set):
            lst = self.str_to_list(value)
            return self.to_type(lst, to_type)

        return self.to_type(value, to_type)

    def translate(self, value):

        if isinstance(value, (tuple, list, set)):
            return (type(value))(self.translate(item) for item in value)

        if not isinstance(value, str):
            return value  # can only perform translations to strings

        # https://docs.python.org/3/library/re.html#re.sub
        for pattern, repl in self.translations:
            value = re.sub(pattern, repl, value)

        return value

    @staticmethod
    def str_to_bool(val):
        """ Interpret string representation of bool values. """
        assert isinstance(val, str)
        try:  # if it is a number
            return float(val) > 0
        except ValueError:
            pass  # process as keywords
        return val.lower() in ('1', 'true', 'yes', 'y', 't')

    # Opinion on str -> list
    #
    # It appears to have become more problematic recently, as the variety
    # of data increase in biothings applications, causing the identifiers
    # and field values hard to escape properly when used in queries.
    #
    # Consider implementing a very safe splitting algorithm that only works
    # for basic cases like a,b,c and use other methods like JSON input to
    # pass in complex queries.

    @staticmethod
    def str_to_list(val):
        """ Cast Biothings-style str to list. """
        try:  # core splitting algorithm
            lst = split_ids(str(val))
        except ValueError as err:
            raise OptionError(str(err))
        return lst

    def str_to_int(self, val):
        """ Convert a numerical string to an integer. """
        assert isinstance(val, str)
        if not self.strict:
            val = self.to_type(val, float)
        return self.to_type(val, int)

    @staticmethod
    def to_type(val, type_):
        """
        Native type casting in Python.
        Fallback approach for type casting.
        """
        try:
            result = (type_)(val)
        except (ValueError, TypeError):
            raise OptionError(f"Expect type {type_.__name__}.")
        return result

class PathArgCvter(Converter):
    """
        Dedicated argument converter for path arguments.
        Correspond to arguments received in tornado for
            RequestHandler.path_args
            RequestHandler.path_kwargs
        See https://www.tornadoweb.org/en/stable/web.html
    """

class QueryArgCvter(Converter):
    """
        Dedicated argument converter for url query arguments.
        Correspond to arguments received in tornado from
            RequestHandler.get_query_argument
        See https://www.tornadoweb.org/en/stable/web.html
    """

    @classmethod
    def str_to_bool(cls, val):
        """ Biothings-style str to bool interpretation """
        # empty string indicates the presence of its key
        # we consider this to be a positive boolean value
        # this is especially useful in url when the user
        # may specify endpoint?op1&op2&op3 in which case
        # it makes sense to consider their keys true.
        return super().str_to_bool(val) or val.lower() == ''

class FormArgCvter(Converter):
    """
        Dedicated argument converter for HTTP body arguments.
        Additionally support JSON seriealization format as values.
        Correspond to arguments received in tornado from
            RequestHandler.get_body_argument
        See https://www.tornadoweb.org/en/stable/web.html
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # If jsoninput evaluates to true,
        # try to load the str as a json dump
        jsoninput = kwargs.get("jsoninput", False)
        if isinstance(jsoninput, bool):
            self.jsoninput = jsoninput
        else:  # it itself can be in argument format
            self.jsoninput = self.convert_to(jsoninput, bool)

    # Opinion on JsonInput
    #
    # The aforementioned feature was partly for convenience and partly
    # for backward compatibility with the original published design.
    #
    # It might be more beneficial to use standard data serialization
    # format like JSON to indicate data type unequivocally.

    def convert_to(self, value, to_type):
        if self.jsoninput:
            try:  # attempt to load as json first
                _value = json.loads(value)
            except json.JSONDecodeError as exc:
                logging.debug(str(exc))
            else:  # no more conversions
                if isinstance(_value, to_type):
                    return _value
        if isinstance(value, to_type):
            return value
        if isinstance(value, str):
            return super().convert_to(value, to_type)
        return self.to_type(value, to_type)

class JsonArgCvter(Converter):
    """
        Dedicated argument converter for JSON HTTP bodys.
        Here it is used for dict JSON objects, with their
        first level keys considered as parameters and
        their values considered as arguments to process.

        May correspond to this tornado implementation:
        https://www.tornadoweb.org/en/stable/web.html#input
    """

    def convert_to(self, value, to_type):

        if isinstance(value, to_type):
            return value   # type matches

        if self.strict:
            # since JSON support value types
            # strict mode enforces it and essentially
            # makes this step a validation step.
            raise OptionError(f"Expect type {to_type.__name__}.")

        return self.to_type(value, to_type)

    def to_type(self, val, type_):

        if issubclass(type_, (list, tuple, set)) and not self.strict:
            val = (val, )  # "abc" -> ["abc"] instead of ["a", "b", "c"]

        return super().to_type(val, type_)


class ReqArgs():

    class Path():

        def __init__(self, args=(), kwargs={}):
            assert isinstance(args, (tuple, list))
            assert isinstance(kwargs, dict)
            self.args = args or ()
            self.kwargs = kwargs or {}

        def __getitem__(self, key):
            try:
                if isinstance(key, int):
                    return self.args[key]
                if isinstance(key, str):
                    return self.kwargs[key]
            except (KeyError, IndexError):
                return None

        def __bool__(self):
            return bool(self.args) or bool(self.kwargs)

        def __str__(self):
            lines = []
            for src in ("args", "kwargs"):
                if getattr(self, src):
                    lines.append("{}={}".format(
                        src, str(getattr(self, src))
                    ))
            return "Path(" + ", ".join(lines) + (")")

    def __init__(self, path=None, query=None, form=None, json_=None):

        assert isinstance(query, (dict, type(None)))
        assert isinstance(form, (dict, type(None)))

        if not isinstance(path, (self.Path, type(None))):
            path = self.Path(*path)

        self.path = path  # positional and named capture group in a routing pattern
        self.query = query  # key value pairs after a question mark at the end of an url
        self.form = form  # type multipart/form-data and application/x-www-form-urlencoded
        self.json = json_ if isinstance(json_, dict) else {}  # type application/json

    def lookup(self, locator, order=None, src=False):

        if isinstance(locator, str):
            locator = Locator(dict(keyword=locator))
        elif isinstance(locator, dict):
            locator = Locator(locator)
        elif not isinstance(locator, Locator):
            raise TypeError("Unknown Locator.")

        if order is None:
            order = ('path', 'query', 'form', 'json')
        elif isinstance(order, str):
            order = (order, )
        elif not isinstance(order, abc.Iterable):
            raise TypeError("Unknown Order.")

        for loc in order:
            try:
                args = getattr(self, loc)
                val = locator.lookin(args)
            except AttributeError:
                _ = "No such location: %s."
                logging.warning(_, loc)
            else:
                if val is not None:
                    return (val, loc) if src else val

        return (None, None) if src else None

    def __str__(self):
        lines = []
        for src in ("path", "query", "form", "json"):
            if getattr(self, src):
                lines.append("{}={}".format(
                    src, str(getattr(self, src))
                ))
        return "ReqArgs(" + ",\n".join(lines) + (")")

class Locator():
    """
        Describes the location of an argument in ReqArgs.
        {
            "keyword": <str>,
            "path": <int or str>,
            "alias": <str or [<str>, ...]>
        }
    """

    def __init__(self, defdict):

        self.keyword = defdict.get('keyword')
        self.path = defdict.get('path')
        aliases = defdict.get('alias', [])

        assert isinstance(self.path, (str, int, type(None)))
        assert isinstance(self.keyword, (str, type(None)))

        if isinstance(aliases, list):
            self.aliases = aliases
        elif isinstance(aliases, str):
            self.aliases = [aliases]
        else:  # validation failed
            raise ValueError("Unknown Alias.")

    @singledispatchmethod
    def lookin(self, location):
        """
        Find an argument in the specified location.
        Use directions indicated in this locator.
        """

    @lookin.register(ReqArgs.Path)
    def _(self, path):
        if self.path is not None:
            return path[self.path]
        return None

    @lookin.register(dict)  # all others
    def _(self, dic):
        if self.keyword in dic:
            return dic[self.keyword]
        for alias in self.aliases:
            if alias in dic:
                return dic[alias]
        return None

class Existentialist():
    """
        Describes the requirement of
        the existance of an argument.
        {
            "default": <object>,
            "required": <bool>,
        }
    """

    def __init__(self, defdict):

        self._defdict = MappingProxyType(defdict)
        self.keyword = defdict.get('keyword')
        self.required = bool(defdict.get('required'))
        self.default = defdict.get('default')

        if self.default and self.required:
            logging.warning(
                ("A default value is set for parameter '%s' "
                 "while 'required' is set to True, making it"
                 "ineffective."), self.keyword)

    def inquire(self, obj):

        if obj is None:
            if self.required:
                raise OptionError(
                    missing=self.keyword,
                    keyword=None,  # empty this field
                    alias=self._defdict.get('alias'))

            obj = self.default

        return obj

class Validator():
    """
        Describes the requirement of
        the existance of an argument.
        {
            "enum": <container>,
            "max": <int>
        }
    """

    def __init__(self, defdict):

        self._defdict = MappingProxyType(defdict)
        self.keyword = defdict.get('keyword')
        self.strict = defdict.get('strict', True)
        self.enum = defdict.get('enum', ())
        self.max = defdict.get('max')

        assert isinstance(self.enum, abc.Container)
        assert isinstance(self.max, (int, type(None)))

    def validate(self, obj):

        if self.enum and not self._in_enum(obj):
            raise OptionError(
                keyword=self.keyword,
                allowed=self.enum,
                alias=self._defdict.get('alias'))

        if self.max:
            if isinstance(obj, (list, tuple, set)):
                self._check_list_max(obj)
            elif isinstance(obj, (int, float, complex)):
                self._check_num_max(obj)

        return obj

    def _in_enum(self, value):

        if isinstance(value, (list, tuple, set)):
            for val in value:
                if not self._in_enum(val):
                    return False
            return True

        return value in self.enum

    def _check_list_max(self, container):

        if len(container) > self.max:
            raise OptionError(
                keyword=self.keyword,
                max=self.max,
                size=len(container)
            )

    def _check_num_max(self, num):

        if isinstance(num, bool):
            return

        if num > self.max:
            raise OptionError(
                keyword=self.keyword,
                max=self.max,
                num=num)


class Option(UserDict):
    """
        A parameter for end applications to consume.
        Find the value of it in the desired *location*.

        For example:
        {
            "keyword": "q",
            "location": ("query", "form", "json"),
            "default": "__all__",
            "type": "str"
        }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._locater = Locator(self)
        self._exists = Existentialist(self)
        self._typify = dict(Converter.subclasses(self))
        self._validator = Validator(self)

        # default argument parsing location order
        self.order = self.get("location", None)
        if self.order == "body":  # shortcut
            self.order = ("form", "json")

    def parse(self, reqargs):

        if not isinstance(reqargs, ReqArgs):
            reqargs = ReqArgs(*reqargs)

        # ------- debug here -------
        # if self.get("keyword") == "q":
        #     print()  # breakpoint
        # ---------------------------

        # find the user input
        val, loc = reqargs.lookup(
            locator=self._locater,
            order=self.order,
            src=True)

        if val is None:
            val = self._exists.inquire(val)
        else:  # type conversion and transform
            val = self._typify[loc].convert(val)
            val = self._typify[loc].translate(val)

        # additional conditions
        val = self._validator.validate(val)

        return val

class OptionSet(UserDict):
    """
        A collection of options that a specific endpoint consumes.
        Divided into *groups* and by the *request methods*.

        For example:
        {
            "*":{"raw":{...},"size":{...},"dotfield":{...}},
            "GET":{"q":{...},"from":{...},"sort":{...}},
            "POST":{"q":{...},"scopes":{...}}
        }
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.groups = set()  # explicit result access groups
        self.setup()  # populate self.optset variable

    def __missing__(self, key):
        self.data[key] = {}
        return self[key]

    def setup(self):
        """
        Apply the wildcard method configurations dict.
        Must call this method after changes to this object.
        """
        # store Option objects used for parse method
        self.optset = defaultdict(dict)

        for method, options in self.data.items():
            for keyword, defdict in options.items():
                option = dict(defdict)
                option["keyword"] = keyword
                self.optset[method][keyword] = Option(option)
                if "group" in option:
                    if isinstance(option["group"], (list, tuple, set)):
                        self.groups.update(option["group"])
                    elif isinstance(option["group"], str):
                        self.groups.add(option["group"])

        wildcards = self.optset.get("*", {})
        for method, options in self.optset.items():
            for keyword, option in wildcards.items():
                options.setdefault(keyword, option)

    def parse(self, method, reqargs):
        """
        Parse a HTTP request, represented by its method and args,
        with this OptionSet and return an attribute dictionary.
        """

        options = self.optset.get(method, self.optset["*"])
        result = defaultdict(dict)  # to accomodate groups

        for keyword, option in options.items():
            try:
                val = option.parse(reqargs)
            except OptionError as err:
                err.info.setdefault("keyword", keyword)
                err.info["alias"] = option.get("alias")
                err.simplify()  # remove empty fields
                raise err  # with helpful info

            if val is not None:
                # TODO: build a new ds for returned result
                if 'group' in option:
                    group = option['group']
                    if isinstance(group, str):
                        result[group][keyword] = val
                    else:  # assume iterable
                        for _group in group:
                            result[_group][keyword] = val
                else:  # top level keywords
                    result[keyword] = val

        # make sure all named groups exist
        for group in self.groups:
            if group not in result:
                result[group] = {}

        return dotdict(result)

class OptionsManager(UserDict):
    """
    A collection of OptionSet(s) that makes up an application.
    Provide an interface to setup and serialize.

    Example:
    {
        "annotation": {"*": {...}, "GET": {...}, "POST": {... }},
        "query": {"*": {...},  "GET": {...}, "POST": {... }},
        "metadata": {"GET": {...}, "POST": {... }}
    }
    """

    def add(self, name, optionset, groups=()):
        if not name:
            logging.warning("Ignore unnamed optionset:\n%s", optionset)
        if name not in self.data:
            self.data[name] = OptionSet(optionset)
            self.data[name].groups.update(groups)
        else:  # update existing optionset
            for method, options in optionset.items():
                # merge second level objects
                self.data[name][method].update(options)
                self.data[name].groups.update(groups)
                self.data[name].setup()  # required

    def log(self):

        res = self._serialize(self)
        return res

    def _serialize(self, obj):
        if isinstance(obj, abc.Mapping):
            _obj = {}
            items = list(obj.items())
            for key, val in sorted(items):
                _obj[key] = self._serialize(val)
            return _obj
        if isinstance(obj, (list, tuple)):
            return [self._serialize(item) for item in obj]
        if isinstance(obj, (str, int)):
            return obj
        if hasattr(obj, "__name__"):
            return obj.__name__
        return str(obj)  # best effort



from os import listdir
from os.path import isfile, join

from tornado.web import RequestHandler
from tornado.template import Template

catalog = Template("""
    <html>
        <head>
            <title> Biothings Hub Logs </title>
        </head>
        <body>
            <h1> Logs </h1>
            <ul>
            {% for log in logs %}
                <li>
                    <a href={{"./" + log }}>
                        {{ log[:-4] if log.endswith('.log') else log }}
                    </a>
                </li>
            {% end %}
            </ul>
        </body>
    </html>
""")

logfile = Template("""
    <html>
        <head>
            <title> {{ "Log " + name }} </title>
        </head>
        <body>
            <h1> {{ name }} </h1>
            <p>
                {% for line in lines %}
                    {{ line }} </br>
                {% end %}
            </p>
        </body>
    </html>
""")

class HubLogHandler(RequestHandler):

    def initialize(self, path):
        self.path = path

    def get(self, filename):

        if not filename:
            logs = sorted([
                f for f in listdir(self.path)
                if isfile(join(self.path, f))
            ])
            self.finish(catalog.generate(logs=logs))
            return

        if not isfile(join(self.path, filename)):
            self.set_status(404)
            return

        with open(join(self.path, filename), 'r') as file:
            self.finish(logfile.generate(
                name=filename,
                lines=file.read().split('\n')
            ))



"""
Elasticsearch Snapshot Feature

Snapshot Config Example:
{
    "cloud": {
        "type": "aws",  # default, only one supported by now
        "access_key": None,
        "secret_key": None,
    },
    "repository": {
        "name": "s3-$(Y)",
        "type": "s3",
        "settings": {
            "bucket": "<SNAPSHOT_BUCKET_NAME>",
            "base_path": "mynews.info/$(Y)",  # per year
            "region": "us-west-2",
        },
        "acl": "private",
    },
    #----------------------------- inferred from build doc from now on
    "indexer": {
        # reference to INDEX_CONFIG
        "env": "local",
    },
    #-----------------------------
    # when creating a snapshot, how long should we wait before querying ES
    # to check snapshot status/completion ? (in seconds)
    "monitor_delay": 60 * 5,
}

SnapshotManager => SnapshotEnvConfig(s)
SnapshotEnvConfig + (Build) -> SnapshotEnv
SnapshotEnv + Index + Snapshot -> SnapshotTaskEnv
"""

import asyncio
import copy
import json
from functools import partial
from pprint import pformat
import time
import biothings.utils.aws as aws
try:
    from biothings import config as btconfig
    from config import logger as logging
except ImportError:
    import sys
    sys.path.insert(1, '/home/biothings/mychem.info/src')
    import config
    from biothings import config_for_app
    config_for_app(config)
    from biothings import config as btconfig
    from config import logger as logging


from biothings.hub import SNAPSHOOTER_CATEGORY, SNAPSHOTMANAGER_CATEGORY
from biothings.hub.databuild.buildconfig import AutoBuildConfig
from biothings.hub.datarelease import set_pending_to_release_note
from biothings.utils.es import ESIndexer
from biothings.utils.es import IndexerException as ESIndexerException
from biothings.utils.hub import template_out
from biothings.utils.hub_db import get_src_build
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager, BaseStatusRegisterer


class BuildSpecificEnv(dict):
    """
    Snapshot Env % Build
    """

    def __init__(self, env_conf, build_doc):
        super().__init__(self._template_out_config(env_conf, build_doc))

    def _template_out_config(self, env_conf, build_doc):
        """
        Template out for special value using build_doc
        Templated values can look like:
            "base_path" : "onefolder/%(_meta.build_version)s"
        where "_meta.build_version" value is taken from build_doc dictionary
        (dot field notation).  In other words, such repo config are dynamic
        and potentially change for each index/snapshot created.
        """
        try:
            strconf = template_out(json.dumps(env_conf), build_doc)
            return json.loads(strconf)
        except Exception as exc:
            logging.exception("Coudn't template out configuration: %s", exc)
            raise

class SnapshotTaskStatusRegister(BaseStatusRegisterer):

    STATUS_NAME = {
        "pre": "pre-snapshot",
        "snapshot": "snapshot",
        "post": "post-snapshot"
    }

    def __init__(self, index_name, snapshot_name, env_conf, build_doc):

        self.index_name = index_name
        self.snapshot_name = snapshot_name
        self.build_doc = build_doc
        self.env_conf = env_conf

        self.logger, self.logfile = get_logger(
            SNAPSHOOTER_CATEGORY, btconfig.LOG_FOLDER)

    @property  # override
    def collection(self):
        return get_src_build()

    def reset_repository_info(self, snapshot_name):
        bdoc = self.build_doc
        if bdoc.get("snapshot", {}).get(snapshot_name):
            if bdoc["snapshot"][snapshot_name].pop("repository", None):
                self.collection.save(bdoc)

    # override
    def register_status(
            self, status, job_info, snapshot_info,
            transient=False, init=False, **extra):

        super().register_status(
            self.build_doc, "snapshot", status,
            transient=transient, init=init, job=job_info,
            snapshot={self.snapshot_name: snapshot_info},
            **extra
        )

    def register_start(self, step):

        return self.register_status(
            status='in progress',
            transient=True, init=True,
            job_info={"step": self.STATUS_NAME[step]},
            snapshot_info={}
        )

    def register_result(self, step, success, res):

        params = {
            'status': 'success' if success else 'failed',
            'job_info': {
                "step": self.STATUS_NAME[step],
            },
            'snapshot_info': {
                "conf": self.env_conf
            }
        }
        if success:
            params['job_info']['result'] = res
            params['snapshot_info'][step] = res
        else:
            params['job_info']['err'] = res
            params['snapshot_info'][step] = None

        return self.register_status(**params)


class SnapshotTaskProcessInfo():
    """
    Generate information about the current process.
    (used to report in the hub)
    """

    def __init__(self, task_env):
        self.source = task_env.index_name
        self.category = SNAPSHOOTER_CATEGORY,
        self.predicates = self.get_predicates()
        self.es_host = task_env.indexer.es_host

    def get_predicates(self):
        return []

    def get_pinfo(self, step, description):
        pinfo = {
            "category": self.category,
            "source": self.source,
            "step": step,
            "description": description
        }
        if self.predicates:
            pinfo["__predicates__"] = self.predicates
        return pinfo

    def get_pinfo_for(self, step):
        step_info = {
            "pre": ("pre-snapshot", None),
            "snapshot": ("snapshot", self.es_host),
            "post": ("post-snapshot", None)
        }
        return self.get_pinfo(*step_info[step])


class SnapshotTaskESStatusInterpreter():

    def __init__(self, res, task_env):

        assert "snapshots" in res,\
            "Can't find snapshot '%s' in repository '%s'" % \
            (task_env.snapshot_name, task_env.repo)

        # assuming only one index in the snapshot,
        # so only check first element
        info = res["snapshots"][0]

        assert "state" in info, \
            "Can't find state in snapshot '%s'" \
            % task_env.snapshot_name

        self.info = info
        self.state = info["state"]
        self.failed_shards = info["shards_stats"]["failed"]

    def is_running(self):
        return self.state in ("INIT", "IN_PROGRESS", "STARTED")

    def is_succeed(self):
        return self.state == "SUCCESS" and self.failed_shards == 0

    def is_partially_succeed(self):
        return self.state == "SUCCESS" and self.failed_shards != 0

class SnapshotTaskEnv():
    """
    SnapshotEnv + Index Name + Snapshot Name
    Access to status register and ES indexer.
    """

    def __init__(self, env, index, snapshot):

        self.env = env
        self.index_name = index
        self.snapshot_name = snapshot
        self.repo = env.env_config['repository']['name']

        self.status = SnapshotTaskStatusRegister(
            index, self.snapshot_name, env.env_config, env.build_doc
        )
        self.indexer = ESIndexer(
            index=index,
            doc_type=env.build_doc['index'][index]['doc_type'],
            es_host=env.build_doc['index'][index]['host'],
            check_index=index is not None
        )

    def snapshot(self):
        return self.indexer.snapshot(self.repo, self.snapshot_name)

    def wait_for_error(self):
        """
        Blocking, execute in a thread.
        """

        while True:
            info = self.indexer.get_snapshot_status(self.repo, self.snapshot_name)
            status = SnapshotTaskESStatusInterpreter(info, self)

            if status.is_running():
                time.sleep(self.env.env_config.get('monitor_delay', 60))

            elif status.is_succeed():
                logging.info(
                    "Snapshot '%s' successfully created (host: '%s', repository: '%s')" %
                    (self.snapshot_name, self.indexer.es_host, self.repo),
                    extra={"notify": True}
                )
                return None

            else:  # failed
                if status.is_partially_succeed():
                    e = ESIndexerException(
                        "Snapshot '%s' partially failed: state is %s but %s shards failed" %
                        (self.snapshot_name, status.state, status.failed_shards))
                else:
                    e = ESIndexerException(
                        "Snapshot '%s' failed: state is %s" %
                        (self.snapshot_name, status.state))

                logging.error(
                    "Failed creating snapshot '%s' (host: %s, repository: %s), state: %s" %
                    (self.snapshot_name, self.indexer.es_host, self.repo, status.state),
                    extra={"notify": True})

                return e


class SnapshotEnv():
    """
    Corresponds to an ES repository for a specific build.
    The repository type can be what are supported by ES.
    """

    def __init__(self, job_manager, env_config, build_doc):

        self.job_manager = job_manager
        self.env_config = env_config  # already build specific
        self.build_doc = build_doc

    def snapshot(self, index, snapshot=None, steps=("pre", "snapshot", "post")):
        """
        Make a snapshot of 'index', with name 'snapshot' in this env.
        Typically used to make a snapshot of an index created by the hub.
        """

        # process params
        if isinstance(steps, str):
            steps = (steps,)
        snapshot = snapshot or index

        # create a task env, so we have access to indexers, etc.
        task_env = SnapshotTaskEnv(self, index, snapshot)

        @asyncio.coroutine
        def do():
            pinfo = SnapshotTaskProcessInfo(task_env)

            # we only allow one repo conf per snapshot name
            task_env.status.reset_repository_info(snapshot)

            funcs = {
                'pre': partial(self.pre_snapshot, task_env),
                'snapshot': partial(self._snapshot, task_env),
                'post': partial(self.post_snapshot, task_env)
            }

            for step in ('pre', 'snapshot', 'post'):
                if step in steps:
                    task_env.status.register_start(step)
                    job = yield from self.job_manager.defer_to_thread(
                        pinfo.get_pinfo_for(step), funcs[step])
                    try:
                        result = yield from job
                    except Exception as e:
                        task_env.status.register_result(step, False, str(e))
                        logging.error("Error running %s.", step)
                        raise
                    else:
                        task_env.status.register_result(step, True, result)
                        logging.info("Step %s done: %s.", step, result)

            set_pending_to_release_note(self.build_doc['_id'])  # TODO: conditional

        return asyncio.ensure_future(do())

    def _snapshot(self, task_env):
        task_env.snapshot()
        logging.info("Snapshot successfully launched.")

        error = task_env.wait_for_error()
        if error:
            raise RuntimeError("Snapshot failed: %s" % error)

        return "created"

    def pre_snapshot(self, task_env):
        pass

    def post_snapshot(self, task_env):
        pass


class SnapshotFSEnv(SnapshotEnv):

    def __init__(self, job_manager, env_config, build_doc):
        super().__init__(job_manager, env_config, build_doc)
        assert env_config['repository']['type'] == 'fs'
        raise NotImplementedError

class SnapshotS3Env(SnapshotEnv):
    """
    Relevent Config Entries:
    {
        "cloud": {
            "type": "aws",  # default, only one supported by now
            "access_key": None,
            "secret_key": None,
        },
        "repository": {
            "name": "s3-$(Y)",
            "type": "s3",
            "settings": {
                "bucket": "<SNAPSHOT_BUCKET_NAME>",
                "base_path": "mynews.info/$(Y)",  # per year
                "region": "us-west-2",
            },
            "acl": "private",
        }
    }
    """

    def __init__(self, job_manager, env_config, build_doc):
        super().__init__(job_manager, env_config, build_doc)
        assert env_config['repository']['type'] == 's3'

    def pre_snapshot(self, task_env):
        """
        Ensure the destination repository is ready.
        Create the bucket and repository if necessary.
        """
        cloud = dict(self.env_config.get("cloud", {}))
        repository = dict(self.env_config.get("repository", {}))

        try:  # check if already created
            task_env.indexer.get_repository(repository["name"])

        except ESIndexerException:
            # first make sure bucket exists
            aws.create_bucket(
                name=repository["settings"]["bucket"],
                region=repository["settings"]["region"],
                aws_key=cloud.get("access_key"),
                aws_secret=cloud.get("secret_key"),
                acl=repository.get("acl", None),  # let aws.create_bucket default it
                ignore_already_exists=True
            )
            logging.info("Create repository:\n%s" % pformat(repository))
            task_env.indexer.create_repository(repository.pop("name"), repository)

        return "repo_ready"


class SnapshotEnvConfig():
    """
    Snapshot Env before Combining with Build Info.
    """

    def __init__(self, name, env_class, env_config):
        self.name = name
        self.env_class = env_class
        self.env_config = env_config

    def get_env(self, job_manager, build_doc=None):
        """
        Get build specific snapshot environment.
        """
        env = BuildSpecificEnv(self.env_config, build_doc or {})
        return self.env_class(job_manager, env, build_doc or {})

class SnapshotManager(BaseManager):

    SNAPSHOT_ENV = {
        "s3": SnapshotS3Env,
        "fs": SnapshotFSEnv
    }

    def __init__(self, index_manager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_manager = index_manager
        self.snapshot_config = {}  # user specified config

    @staticmethod
    def pending_snapshot(build_name):
        src_build = get_src_build()
        src_build.update({"_id": build_name}, {"$addToSet": {"pending": "snapshot"}})

    @staticmethod
    def get_build_doc(index_name):
        src_build = get_src_build()
        doc = src_build.find_one({"index." + index_name: {"$exists": True}})
        if not doc:
            logging.error("No build associated with index %s.", index_name)
        return doc

    # override
    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            for job in build.get("jobs", []):
                if job.get("status", "") == "in progress":
                    logging.warning(
                        "Found stale build '%s', marking snapshot status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
            src_build.replace_one({"_id": build["_id"]}, build)

    # override
    def poll(self, state, func):
        super().poll(state, func, col=get_src_build())

    def configure(self, snapshot_confdict):
        """
        Configure manager with snapshot config dict.
        See SNAPSHOT_CONFIG in config_hub.py for the format.
        """
        self.snapshot_config = copy.deepcopy(snapshot_confdict)
        for env, envconf in self.snapshot_config.get("env", {}).items():
            if envconf.get("cloud"):
                assert envconf["cloud"]["type"] == "aws", \
                    "Only Amazon AWS cloud is supported at the moment"
            repo_type = envconf.get("repository", {}).get("type")
            if not repo_type:
                raise ValueError("Repository type not specified.")
            if repo_type not in self.SNAPSHOT_ENV.keys():
                raise ValueError("Unsupported repository type %s.", repo_type)
            try:
                self.register[env] = SnapshotEnvConfig(
                    name=env,
                    env_class=self.SNAPSHOT_ENV[repo_type],
                    env_config=envconf
                )
            except Exception as e:
                logging.exception(
                    "Couldn't setup snapshot environment '%s' because: %s" %
                    (env, e))

    def snapshot(self, env, index, snapshot=None,
                 steps=("pre", "snapshot", "post")):
        """
        Create a snapshot named "snapshot" (or, by default, same name as the index)
        from "index" according to environment definition (repository, etc...) "env".
        """
        if env not in self.register:
            raise ValueError("Unknown snapshot environment '%s'." % env)
        build_doc = self.get_build_doc(index)
        if not build_doc:
            logging.warning("The index is not created by the hub.")
        env_for_build = self[env].get_env(self.job_manager, build_doc)
        return env_for_build.snapshot(index, snapshot=snapshot, steps=steps)

    def snapshot_build(self, build_doc):
        """
        Create a snapshot basing on the autobuild settings in the build config.
        If the build config associated with this build has:
        {
            "autobuild": {
                "type": "snapshot", // implied when env is set. env must be set.
                "env": "local" // which es env to make the snapshot.
            },
            ...
        }
        Attempt to make a snapshot for this build on the specified es env "local".
        """
        @asyncio.coroutine
        def _():
            autoconf = AutoBuildConfig(build_doc['build_config'])
            env = autoconf.auto_build.get('env')
            assert env, "Unknown autobuild env."
            try:
                latest_index = list(build_doc['index'].keys())[-1]
            except Exception:
                logging.info("No index already created, now create one.")
                yield from self.index_manager.index(env, build_doc['_id'])
                latest_index = build_doc['_id']
            return self.snapshot(env, latest_index)
        return asyncio.ensure_future(_())

    def snapshot_info(self, env=None, remote=False):
        return copy.deepcopy(self.snapshot_config)


def test():
    from biothings.utils.manager import JobManager
    from biothings.hub.dataindex.indexer import IndexManager
    loop = asyncio.get_event_loop()
    job_manager = JobManager(loop)
    index_manager = IndexManager(job_manager=job_manager)
    index_manager.configure(config.INDEX_CONFIG)
    snapshot_manager = SnapshotManager(
        index_manager=index_manager,
        job_manager=job_manager,
        poll_schedule="* * * * * */10"
    )
    snapshot_manager.configure(config.SNAPSHOT_CONFIG)
    # snapshot_manager.poll("snapshot",snapshot_manager.snapshot_build)

    async def test_code():
        snapshot_manager.snapshot('prod', 'mynews_202009170234_fjvg7skx', steps="post")

    asyncio.ensure_future(test_code())
    loop.run_forever()


if __name__ == '__main__':
    test()



"""
    Biothings Test Helper

    Envs:

    TEST_SCHEME
    TEST_PREFIX
    TEST_HOST
    TEST_CONF

"""
import os
from functools import partial

import pytest
import requests
from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase

from biothings.web.settings import BiothingESWebSettings


class BiothingsDataTest():

    # relative path parsing configuration
    scheme = 'http'
    prefix = 'v1'
    host = ''

    def request(self, path, method='GET', expect=200, **kwargs):
        """
        Use requests library to make an HTTP request.
        Ensure path is translated to an absolute path.
        Conveniently check if status code is as expected.
        """
        url = self.get_url(path)
        res = requests.request(method, url, **kwargs)
        assert res.status_code == expect, res.text

        return res

    def get_url(self, path):
        """
        Try best effort to get a full url to make a request.
        Return an absolute url when class var 'host' is defined.
        If not, return a path relative to the host root.
        """
        scheme = os.getenv("TEST_SCHEME", self.scheme)
        prefix = os.getenv("TEST_PREFIX", self.prefix).strip('/')
        host = os.getenv("TEST_HOST", self.host).strip('/')

        # already an absolute path
        if path.lower().startswith(("http://", "https://")):
            return path

        # path standardization
        if not path.startswith('/'):
            if prefix:  # append prefix
                path = '/'.join((prefix, path))
            path = '/' + path

        # host standardization
        if host:
            path = f"{scheme}://{host}{path}"

        return path

    def query(self, method='GET', endpoint='query', hits=True, data=None, json=None, **kwargs):
        """
        Make a query and assert positive hits by default.
        Assert zero hit when hits is set to False.
        """

        if method == 'GET':
            res = self.request(
                endpoint, params=kwargs,
                data=data, json=json).json()

            assert bool(res.get('hits')) == hits
            return res

        if method == 'POST':
            res = self.request(
                endpoint, method=method,
                params=kwargs, data=data,
                json=json).json()

            for item in res:  # list
                if "_id" not in item:
                    _hits = False
                    break
            else:
                _hits = bool(res)
            assert _hits is hits
            return res

        raise ValueError('Invalid Request Method.')

    @staticmethod
    def msgpack_ok(packed_bytes):
        """ Load msgpack into a dict """
        try:
            import msgpack
        except ImportError:
            pytest.skip('Msgpack is not installed.')
        try:
            dic = msgpack.unpackb(packed_bytes)
        except BaseException:  # pylint: disable=bare-except
            assert False, 'Not a valid Msgpack binary.'
        return dic


class BiothingsWebAppTest(BiothingsDataTest, AsyncHTTPTestCase):
    """
        Starts the tornado application to run tests locally.
        Need a config.py under the current working dir.
    """

    @classmethod
    def setup_class(cls):
        conf = os.getenv("TEST_CONF", 'config')
        cls.settings = BiothingESWebSettings(conf)
        prefix = cls.settings.API_PREFIX
        version = cls.settings.API_VERSION
        cls.prefix = f'{prefix}/{version}'

    # override
    def get_new_ioloop(self):
        return IOLoop.current()

    # override
    def get_app(self):
        return self.settings.get_app()

    # override
    def request(self, path, method="GET", expect=200, **kwargs):

        url = self.get_url(path)

        func = partial(requests.request, method, url, **kwargs)
        res = self.io_loop.run_sync(
            lambda: self.io_loop.run_in_executor(None, func))

        assert res.status_code == expect, res.text
        return res

    # override
    def get_url(self, path):

        path = BiothingsDataTest.get_url(self, path)
        return AsyncHTTPTestCase.get_url(self, path)


# Compatibility
BiothingsTestCase = BiothingsWebAppTest




import importlib.util
import inspect
import logging
import os
import types
from importlib import import_module
from pprint import pformat
from pydoc import locate

import tornado.log
from tornado.ioloop import IOLoop
from tornado.web import Application

from biothings.utils.web.userquery import ESUserQuery
from biothings.web.handlers import BaseAPIHandler, BaseESRequestHandler
from biothings.web.options import OptionsManager as OptionSets
from biothings.web.utils import DevInfo, FieldNote

from . import default as web_default
from .data import DataConnections, DataMetadata, DataPipeline

try:
    from raven.contrib.tornado import AsyncSentryClient
except ImportError:
    __SENTRY_INSTALLED__ = False
else:
    __SENTRY_INSTALLED__ = True

class BiothingConfigError(Exception):
    pass

class BiothingWebSettings():
    '''
    A container for the settings that configure the web API.

    * Environment variables can override settings of the same names.
    * Default values are defined in biothings.web.settings.default.

    '''
    _default = object()
    _parent = object()
    _user = object()

    def __init__(self, config=None, parent=None, **kwargs):
        '''
        :param config: a module that configures this biothing
            or its fully qualified name,
            or its module file path.
        '''
        self._default = web_default
        self._parent = parent  # config package
        self._user = self.load_module(config, object())
        self.logger.info("%s", self._user)  # log file location

        # process keyword setting override
        for key, value in kwargs.items():
            setattr(self._user, key, value)

        # process environment variable override of named settings
        for name in os.environ:
            if hasattr(self, name):
                new_value = None
                if isinstance(getattr(self, name), str):
                    new_value = os.environ[name]
                elif isinstance(getattr(self, name), bool):
                    new_value = os.environ[name].lower() in ('true', '1')
                if new_value is not None:
                    self.logger.info("$ %s = %s", name, os.environ[name])
                    setattr(self._user, name, new_value)
                else:  # cannot override dict, array, object type...
                    self.logger.error("Env %s is not suppored.", name)

        # ----------- #
        self.validate()
        # ----------- #

        self.fieldnote = FieldNote(self.AVAILABLE_FIELDS_NOTES_PATH)
        self.devinfo = DevInfo(self.APP_GIT_REPOSITORY)

        # initialize payload for standalone tracking batch
        self.tracking_payload = []

        self.optionsets = OptionSets()
        self.handlers = {}

    @staticmethod
    def load_module(config, default=None):
        """
        Ensure config is a module.
        If config does not evaluate,
        Return default if it's provided.
        """
        if isinstance(config, types.ModuleType):
            return config
        elif isinstance(config, str) and config.endswith('.py'):
            spec = importlib.util.spec_from_file_location("config", config)
            config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(config)
            return config
        elif isinstance(config, str) and config:
            return import_module(config)
        elif not config and default:
            return default
        raise BiothingConfigError()

    def __getattr__(self, name):
        if hasattr(self._user, name):
            return getattr(self._user, name)
        elif hasattr(self._parent, name):
            return getattr(self._parent, name)
        elif hasattr(self._default, name):
            return getattr(self._default, name)
        else:  # not provided and no default
            raise AttributeError(name)

    @property
    def logger(self):
        return logging.getLogger('biothings.web.settings')

    @staticmethod
    def load_class(kls):
        """
        Ensure config is a module.
        If config does not evaluate,
        Return default if it's provided.
        """
        if inspect.isclass(kls):
            return kls
        if isinstance(kls, str):
            return locate(kls)
        raise BiothingConfigError()

    def _generate_app_settings(self, override=None):
        """
        Generates settings for tornado.web.Application. This result and the
        method below can define a tornado application to start a web server.
        """
        settings = {
            'biothings': self,
            'autoreload': False,
            'debug': False,
        }
        settings.update(override or {})
        supported_keywords = (
            'default_handler_class', 'default_handler_args', 'template_path',
            'log_function', 'compress_response', 'cookie_secret',
            'login_url', 'static_path', 'static_url_prefix')

        for setting in supported_keywords:
            if hasattr(self, setting.upper()):
                settings[setting] = getattr(self, setting.upper())

        if __SENTRY_INSTALLED__ and self.SENTRY_CLIENT_KEY:
            # Setup error monitoring with Sentry. More on:
            # https://docs.sentry.io/clients/python/integrations/#tornado
            settings['sentry_client'] = AsyncSentryClient(self.SENTRY_CLIENT_KEY)

        return settings

    def _generate_app_handlers(self, addons=None):
        '''
        Generates the tornado.web.Application `(regex, handler_class, options) tuples
        <http://www.tornadoweb.org/en/stable/web.html#application-configuration>`_.
        '''
        handlers = {}
        addons = addons or []
        for rule in self.APP_LIST + addons:
            pattern = rule[0]
            handler = self.load_class(rule[1])
            setting = rule[2] if len(rule) == 3 else {}
            assert handler, rule[1]
            if issubclass(handler, BaseAPIHandler) and handler.name:
                handler_name = handler.name
                handler_options = handler.kwargs
                setting_attr = '_'.join((handler_name, 'kwargs')).upper()
                setting_options = getattr(self, setting_attr, {})
                self.optionsets.add(handler_name, handler_options, handler.kwarg_groups)
                self.optionsets.add(handler_name, setting_options)
            if '{typ}' in pattern:
                if not issubclass(handler, BaseESRequestHandler):
                    raise BiothingConfigError()
                for biothing_type in self.BIOTHING_TYPES:
                    _pattern = pattern.format(
                        pre=self.API_PREFIX,
                        ver=self.API_VERSION,
                        typ=biothing_type).replace('//', '/')
                    _setting = dict(setting)
                    _setting['biothing_type'] = biothing_type
                    handlers[_pattern] = (_pattern, handler, _setting)
            elif '{pre}' in pattern or '{ver}' in pattern:
                pattern = pattern.format(
                    pre=self.API_PREFIX,
                    ver=self.API_VERSION).replace('//', '/')
                if '()' not in pattern:
                    handlers[pattern] = (pattern, handler, setting)
            else:  # no pattern translation
                handlers[pattern] = (pattern, handler, setting)

        self.handlers = handlers
        handlers = list(handlers.values())
        self.logger.info('API Handlers:\n%s', pformat(handlers, width=200))
        return handlers

    def get_app(self, settings=False, handlers=None):
        """
        Return the tornado.web.Application defined by this settings.
        This is primarily how an HTTP server interacts with this class.
        Additional settings and handlers accepted as parameters.
        """
        # config package
        if self._user.__name__ == self._user.__package__:
            attrs = [getattr(self._user, attr) for attr in dir(self._user)]
            confs = [attr for attr in attrs if isinstance(attr, types.ModuleType)]
            _settings = [self.__class__(_attr, self._user) for _attr in confs]
            _handlers = [(f'/{c.API_PREFIX}/.*', c.get_app(settings)) for c in _settings]
            _handlers += handlers or []  # second level front pages won't be exposed
        else:  # config module
            _handlers = self._generate_app_handlers(handlers)
        _settings = self._generate_app_settings(settings)
        return Application(_handlers, **_settings)

    def configure_logger(self, logger):
        '''
        Configure a logger's formatter to use the format defined in this web setting.
        '''
        try:
            if self.LOGGING_FORMAT and logger.hasHandlers():
                for handler in logger.handlers:
                    if isinstance(handler.formatter, tornado.log.LogFormatter):
                        handler.formatter._fmt = self.LOGGING_FORMAT
        except Exception:
            self.logger.exception('Error configuring logger %s.', logger)

    def validate(self):
        '''
        Validate the settings defined for this web server.
        '''
        assert self.API_VERSION or self.API_PREFIX
        assert isinstance(self.LIST_SIZE_CAP, int)
        assert isinstance(self.ACCESS_CONTROL_ALLOW_METHODS, str)
        assert isinstance(self.ACCESS_CONTROL_ALLOW_HEADERS, str)

        self.ES_INDICES = dict(self.ES_INDICES)
        self.ES_INDICES[self.ES_DOC_TYPE] = self.ES_INDEX
        self.BIOTHING_TYPES = list(self.ES_INDICES.keys())

class BiothingESWebSettings(BiothingWebSettings):
    """
    With additional settings pecific to an elasticsearch backend.
    """

    def __init__(self, config=None, parent=None, **kwargs):

        super(BiothingESWebSettings, self).__init__(config, parent, **kwargs)

        # user query data
        self.userquery = ESUserQuery(self.USERQUERY_DIR)

        self.connections = DataConnections(self)
        self.metadata = DataMetadata(self)
        self.pipeline = DataPipeline(self)

        IOLoop.current().add_callback(self._initialize)

    async def _initialize(self):

        # failures will be logged concisely
        logging.getLogger('elasticsearch.trace').propagate = False

        await self.connections.log_versions()

        # populate source mappings
        for biothing_type in self.ES_INDICES:
            await self.metadata.refresh(biothing_type)

        # resume normal log flow
        logging.getLogger('elasticsearch.trace').propagate = True

    def validate(self):
        '''
        Additional ES settings to validate.
        '''
        super().validate()

        assert isinstance(self.ES_INDEX, str)
        assert isinstance(self.ES_DOC_TYPE, str)
        assert isinstance(self.ES_INDICES, dict)
        assert '*' not in self.ES_DOC_TYPE



"""
    Data utility functions for web settings.
    Most of the classes are specific to elasticsearch.
    Typically one instance of each per settings class.
"""

from collections import defaultdict
from datetime import datetime
from functools import reduce
from operator import add

import elasticsearch
import elasticsearch_dsl
from elasticsearch_dsl.connections import Connections
from dateutil.parser import parse as dtparse

from biothings.utils.web.es import get_es_versions
from biothings.utils.web.es_transport import (BiothingsAsyncTransport,
                                              BiothingsTransport)
from biothings.utils.web.run import run_once

__all__ = [
    'DataConnections',
    'DataPipeline',
    'DataMetadata',
]


class DataConnections:
    """
        Elasticsearch Database Connection
        Connections are created upon first call
    """
    _should_log_es_py_ver = run_once()  # evaluate to true once
    _should_log_es_host_ver = run_once()  # evaluate to true once per host

    def __init__(self, settings):

        self.settings = settings
        self._connections = Connections()

        connection_settings = {
            "hosts": self.settings.ES_HOST,
            "timeout": self.settings.ES_CLIENT_TIMEOUT
        }
        connection_settings.update(transport_class=BiothingsTransport)
        self._connections.create_connection(alias='sync', **connection_settings)
        connection_settings.update(transport_class=BiothingsAsyncTransport)
        if self.settings.ES_SNIFF:
            connection_settings.update(sniffer_timeout=60)
            connection_settings.update(sniff_on_start=True)
            connection_settings.update(sniff_on_connection_fail=True)
            connection_settings.update(retry_on_timeout=True)
        self._connections.create_connection(alias='async', **connection_settings)

    async def log_versions(self):

        if DataConnections._should_log_es_py_ver():

            self.settings.logger.info(
                "Python Elasticsearch Version: %s",
                '.'.join(map(str, elasticsearch.__version__)))
            self.settings.logger.info(
                "Python Elasticsearch DSL Version: %s",
                '.'.join(map(str, elasticsearch_dsl.__version__)))

            if elasticsearch.__version__[0] != elasticsearch_dsl.__version__[0]:
                self.settings.logger.error(
                    "ES Pacakge Version Mismatch with ES-DSL.")

        if DataConnections._should_log_es_host_ver(self.settings.ES_HOST):

            versions = await get_es_versions(self.async_client)
            self.settings.logger.info(
                '[%s] %s: %s', self.settings.ES_HOST,
                versions["elasticsearch_cluster"],
                versions["elasticsearch_version"])

            major_version = versions["elasticsearch_version"].split('.')[0]
            if major_version.isdigit() and int(major_version) != elasticsearch.__version__[0]:
                self.settings.logger.error(
                    "ES Python Version Mismatch.")

    def get_connection(self, connection):
        return self._connections.get_connection(connection)

    @property
    def client(self):
        '''
        Return the blocking elasticsearch client.
        '''
        return self.get_connection('sync')

    @property
    def async_client(self):
        '''
        Return the async elasitcsearch client.
        API calls return awaitable objects.
        '''
        return self.get_connection('async')

class DataPipeline:

    def __init__(self, settings):

        _load_class = settings.load_class
        self.query_builder = _load_class(settings.ES_QUERY_BUILDER)(settings)
        self.query_backend = _load_class(settings.ES_QUERY_BACKEND)(settings)
        self.result_transform = _load_class(settings.ES_RESULT_TRANSFORM)(settings)

    def build(self, *args, **kwargs):
        return self.query_builder.build(*args, **kwargs)

    def execute(self, *args, **kwargs):
        return self.query_backend.execute(*args, **kwargs)

    def transform(self, *args, **kwargs):
        return self.result_transform.transform(*args, **kwargs)

    def transform_mapping(self, *args, **kwargs):
        return self.result_transform.transform_mapping(*args, **kwargs)

class DataMetadata:

    def __init__(self, settings):

        # defined in configs
        self._biothing_default = settings.ES_DOC_TYPE
        self._biothing_indices = settings.ES_INDICES

        # defined in biothings.web.settings
        self.client = settings.connections.async_client
        self.logger = settings.logger

        # cached, generated from mappings
        self.biothing_metadata = defaultdict(dict)
        self.biothing_mappings = defaultdict(dict)
        self.biothing_licenses = defaultdict(dict)

    async def refresh(self, biothing_type=None):
        """
        Read ES index mappings for the corresponding biothing_type,
        Populate datasource info and field properties from mappings.
        """
        if not biothing_type:
            biothing_type = self._biothing_default
        try:
            info = await self.client.indices.get(
                index=self._biothing_indices[biothing_type]
            )
            count = await self.client.count(
                index=self._biothing_indices[biothing_type]
            )
        except elasticsearch.TransportError as exc:
            self.logger.error('Error loading [%s] indices.', biothing_type)
            self.logger.debug(str(exc))
            return None

        reader = BiothingMetadataReader(biothing_type, info, count)
        self.biothing_metadata[biothing_type] = reader.get_metadata()
        self.biothing_mappings[biothing_type] = reader.get_mappings()
        self.biothing_licenses[biothing_type] = reader.get_licenses()

        return info  # raw index info


class BiothingMetadataReader:
    """
    Read http://<elasticsearch>/<index_pattern>/ and ./_stats
    If the pattern matches one index, then that index's info will be used.
    If the pattern matches multiple indices, then the results will be combined.
    If the pattern matches no index, then empty dictionaries are returned.
    """

    def __init__(self, biothing_type, info, count):

        self.biothing_type = biothing_type
        self.document_count = count

        self.indices_info = {
            index: ESIndex(biothing_type, **index_info)
            for index, index_info in info.items()
        }

    def get_mappings(self):
        """
        Mapping properties used for metadata field endpoint. For example:
        {
            'aeolus': {'properties': {'drug_code': {...}, ...}},
            'cadd': {'properties': {'1000g': {...}, 'alt': {...}, ...}},
            'cgi': {'properties': {'association': {...}, ...}},
            'chebi': {'properties': {'brand_names': {...}, ...}},
            ...
        }
        """
        mappings = (info.get_mappings() for info in self.indices_info.values())
        mappings = reduce(add, mappings).to_dict() if mappings else {}
        return mappings

    def get_licenses(self):
        """
        Source-URL pairs that contains the data licencing information. Example:
        {
            'aeolus': 'http://bit.ly/2DIxWwF',
            'cadd': 'http://bit.ly/2TIuab9',
            'cgi': 'http://bit.ly/2FqS871',
            'chebi': 'http://bit.ly/2KAUCAm',
            ...
        }
        """
        licenses = (info.get_licenses() for info in self.indices_info.values())
        licenses = reduce(add, licenses).to_dict() if licenses else {}
        return licenses

    def get_metadata(self):
        """
        Provide description about the data under this type. Example:
        {
            "biothing_type": "variant",
            "build_date": "2020-05-08T17:33:59.756164",
            "build_version": "20200508",
            "src": {"cosmic": { ... }, ... },
            "stats": {"total": 928585791 }
        }
        """
        metadata = (info.get_metadata() for info in self.indices_info.values())
        metadata = reduce(add, metadata).to_dict() if metadata else {}
        if metadata:
            if metadata.get('biothing_type') == "__multiple__":
                metadata['biothing_type'] = self.biothing_type
            metadata['stats']['total'] = self.document_count['count']
        metadata['_biothing'] = self.biothing_type
        metadata['_indices'] = list(self.indices_info.keys())
        return metadata


class ESIndex:
    """
    Read one index's info http://<elasticsearch>/<index>.
    Return combinable BiothingMetaProp objects.
    """

    def __init__(self, biothing, aliases, mappings, settings):
        self.biothing = biothing
        self.aliases = aliases
        self.mappings = ESIndexMappings(mappings)
        self.settings = ESIndexSettings(settings)

    def get_metadata(self):
        """
        Return BiothingHubMetadata instance.
        Populate empty metadata basing on index settings.
        Fill in empty stats field if not provided.
        """
        if self.mappings.metadata:
            try:
                return BiothingHubMeta(**self.mappings.metadata)
            except KeyError:
                pass
        return BiothingHubMeta(
            biothing_type=self.biothing,
            build_date=self.settings.get_creation_date().isoformat(),
            build_version=self.settings.get_index_version(),
            src={},
            stats={}
        )

    def get_licenses(self):
        return BiothingLicenses(
            self.mappings.extract_licenses()
        )

    def get_mappings(self):
        return BiothingMappings(
            self.mappings.properties
        )


class ESIndexSettings:
    """
    Object representation of ES index settings.
    {
        "index": {
            "number_of_shards": "1",
            "auto_expand_replicas": "0-1",
            "provided_name": ".tasks",
            "creation_date": "1566293197607",
            "priority": "2147483647",
            "number_of_replicas": "0",
            "uuid": "yWBk0qw0QXmEuxJFas3mIg",
            "version": {
                "created": "6050099"
            }
        }
    }
    """

    def __init__(self, setting):
        self.index = setting['index']

    def get_creation_date(self):
        return datetime.fromtimestamp(int(self.index['creation_date'])/1000)

    def get_index_version(self):
        if 'updated' in self.index['version']:
            return self.index['version']['updated']
        return self.index['version']['created']


class ESIndexMappings:
    """
    Object representation of ES index mappings:
    {
        # this level is only available for es6
        "<doc_type> : {
            'properties': { ... },  ---> mapping
            '_meta': {
                "src" : { ... }     ---> licenses
                ...
            },              -----------> metadata
            ...
        }
    }
    """

    def __init__(self, mapping):
        # assume package version corresponds to that of es
        if mapping and elasticsearch.__version__[0] < 7:
            # remove doc_type, support 1 type per index
            mapping = next(iter(mapping.values()))
        self.enabled = mapping.pop('enabled', True)
        self.dynamic = mapping.pop('dynamic', True)
        self.properties = mapping.get('properties', {})
        self.metadata = mapping.get('_meta', {})

    def extract_licenses(self):
        """
        Return source name - license url pairs.
        """
        licenses = {}
        for src, info in self.metadata.get('src', {}).items():
            if 'license_url_short' in info:
                licenses[src] = info['license_url_short']
            elif 'license_url' in info:
                licenses[src] = info['license_url']
        return licenses

class BiothingMetaProp:

    def __add__(self, other):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError

class BiothingLicenses(BiothingMetaProp):

    def __init__(self, licenses):
        self.licenses = licenses

    def __add__(self, other):
        # TODO log conflicts
        licenses = dict(self.licenses)
        licenses.update(other.licenses)
        return BiothingLicenses(licenses)

    def to_dict(self):
        return dict(self.licenses)


class BiothingMappings(BiothingMetaProp):

    def __init__(self, properties):
        self.properties = properties

    def __add__(self, other):
        # TODO conflicting fields
        mappings = dict(self.properties)
        mappings.update(other.properties)
        return BiothingMappings(mappings)

    def to_dict(self):
        return dict(self.properties)


class BiothingHubMeta(BiothingMetaProp):

    def __init__(self, **metadata):  # dict

        self.biothing_type = metadata.get('biothing_type')
        # self.build_date = datetime.fromisoformat(metadata['build_date']) # python3.7 syntax
        # self.build_date = datetime.strptime(metadata['build_date'], "%Y-%m-%dT%H:%M:%S.%f")
        self.build_date = dtparse(metadata['build_date']).astimezone()    # this handles with timestamp string w/wo timezone
        self.build_version = metadata.get('build_version')
        self.src = metadata.get('src', {})
        self.stats = metadata.get('stats', {})

    def to_dict(self):

        return {
            'biothing_type': self.biothing_type,
            'build_date': self.build_date.isoformat(),
            'build_version': self.build_version,
            'src': self.src,
            'stats': self.stats
        }

    def __add__(self, other):

        # combine biothing_type field
        biothing_type = self.biothing_type,
        if other.biothing_type != self.biothing_type:
            biothing_type = '__multiple__'

        # take the latest build_date
        build_date = self.build_date
        if other.build_date > self.build_date:
            build_date = other.build_date

        # combine build_version field
        build_version = self.build_version
        if other.build_version != build_version:
            build_version = '__multiple__'

        # combine source field
        src = dict(self.src)
        src.update(other.src)

        # add up stats field
        stats = dict(self.stats)
        for key, value in other.stats.items():
            if key in stats:
                stats[key] += value
            else:  # new key
                stats[key] = value

        return BiothingHubMeta(
            biothing_type=biothing_type,
            build_date=build_date.isoformat(),
            build_version=build_version,
            src=src,
            stats=stats)




import json
import os

from biothings.utils.version import get_software_info
from biothings.utils.web.es import get_es_versions


class FieldNote:

    def __init__(self, path):

        try:  # populate field notes if exist
            inf = open(path, 'r')
            self._fields_notes = json.load(inf)
            inf.close()
        except Exception:
            self._fields_notes = {}

    def get_field_notes(self):
        '''
        Return the cached field notes associated with this instance.
        '''
        return self._fields_notes

class DevInfo:

    def __init__(self, path):

        # for metadata dev details
        if os.path.isdir(os.path.join(path, '.git')):
            self._git_repo_path = path
        else:
            self._git_repo_path = None

    def get_git_repo_path(self):

        return self._git_repo_path
